#include <stdio.h>
#include <sys/time.h>

#include "fft.h"
#include "constant.h"

/*					    */
/*	Precision dependant declarations    */
/*					    */
#ifdef	ZOMPLEX

#define TOLERANCE   DTOLERANCE
typedef	    double	this_half;
typedef	    zomplex	this_type;

#define	    THIS_SUM	dsum_
#define	    THIS_GEN	zgen_
#define	    THIS_FT	zft1d_

#define	    THIS_FFTI	zfft1di
#define	    THIS_FFT	zfft1d
#define	    THIS_SCAL	zscal1d

#endif
#ifdef	COMPLEX

typedef	    float	    this_half;
typedef	    complex	    this_type;

#define TOLERANCE   STOLERANCE

#define	    THIS_SUM	ssum_
#define	    THIS_GEN	cgen_
#define	    THIS_FT	cft1d_

#define	    THIS_FFTI	cfft1di
#define	    THIS_FFT	cfft1d
#define	    THIS_SCAL	cscal1d

#endif

/*					    */
this_half THIS_SUM();

void inimat_();
void GetArguments();
void get_values();

int size, inc, n_trials;
this_type *pa, *pb, *pRef, *pWork;

main(argc,argv)
int argc;
char *argv[];
{
    int i, j, k, n_total, is_wrong, nerr;
    double x, y, dx, dy, emax, sum, err, maxerr;
    this_half	t;

/* ******************************************************* */
    GetArguments( argc, argv);
/* ******************************************************* */

    srandom( (123*getpid()) | 0x01);

    for( k = 0; k < n_trials ; k++) { 
	get_values();

	if( !(k%1) && (k)) printf("\n");
	printf( "<%3d,%3d>", size, inc);
	fflush(stdout);

	n_total = ((size+1)*inc);
	pa = (this_type *)malloc( (3 * n_total + size + FACTOR_SPACE) * sizeof(this_type));
	if( !(pa) ) {
	    fprintf( stderr, "Could not allocate ... Exiting\n");
	    exit (-1);
	}
	pb = pa + n_total;
	pRef = pb + n_total;
	pWork = pRef + n_total;

/* *******************************************************
	Initialize arrays
******************************************************* */
	THIS_GEN(pRef, &n_total);
	bcopy( pRef, pa, n_total*sizeof(this_type));
	bcopy( pRef, pb, n_total*sizeof(this_type));

/* *******************************************************
	direct Fourier Transform
******************************************************* */
	    j = -1;
	THIS_FT( &j, &size, pb, &inc);
	pWork = THIS_FFTI( size, pWork);
	is_wrong = THIS_FFT( -1, size, pa, inc, pWork);

	emax = TOLERANCE*sqrt(size);
	for(i = 0, nerr=0 ; i < (size*inc) ; i ++) {
	    x = pa[i].re - pb[i].re;
	    y = pa[i].im - pb[i].im;
	    if( (t= x*x + y*y) > (emax)) {
		fprintf( stderr, "@", size, i, t);
		nerr++;
		break;
	    }
	}
	if( !(nerr)){
		fprintf( stderr, ".", size, i, t);
	}
	i = inc * 2;
	x =  THIS_SUM( &size, pRef, &i);
	y =  THIS_SUM( &size, ((char *)pRef) + sizeof(this_half), &i);
	dx = x - pa[0].re;
	dy = y - pa[0].im;
	if( (t= dx *dx + dy*dy) > TOLERANCE){
		fprintf( stderr, 
		"\nError direct FFT(%d) at index #0 : (%f,%f)<->(%f,%f)error(%f,%f)",
		size, pa[0].re, pa[0].im,x,y, dx, dy);

	}
	dx = x - pb[0].re;
	dy = y - pb[0].im;
	if( (t= dx *dx + dy*dy) > TOLERANCE){
		fprintf( stderr, 
		"\nError direct DFT(%d) at index #0 : (%f,%f)<->(%f,%f)error(%f,%f)\n",
		size, pa[0].re, pa[0].im,x,y, dx, dy);

	}
/* *******************************************************
	inverse Fourier Transform
******************************************************* */
	is_wrong = THIS_FFT( 1, size, pa, inc, pWork);

	i = inc * 2;
	x =  THIS_SUM( &size, pb, &i);
	y =  THIS_SUM( &size, ((char *)pb) + sizeof(this_half), &i);
	dx = x - pa[0].re;
	dy = y - pa[0].im;
	nerr = 0;
	emax = size * TOLERANCE;
	if( (t= dx *dx + dy*dy) > TOLERANCE){
		fprintf( stderr, "#");
		nerr ++;
	}
	if( !(nerr)){
		fprintf( stderr, ".", size, i, t);
	}

	t = 1. / (double)size;
	THIS_SCAL( size, t, pa, inc);

	emax = TOLERANCE;
	emax = emax * emax;

	sum = 0.;
	err= 0.;
	nerr= 0;
	maxerr= 0.;
	for(i = 0 ; i < (size*inc) ; i += inc) {
	    sum += (pRef[i].re * pRef[i].re) + (pRef[i].im * pRef[i].im);
	    x = pa[i].re - pRef[i].re;
	    y = pa[i].im - pRef[i].im;
	    if( (t= x*x + y*y) > (emax))
		nerr++;
	    err += t;
	    if( t > maxerr) maxerr = t;
	}
	err = sqrt(err / (double)size);
	sum = sqrt(sum / (double)size);
	maxerr = sqrt(maxerr);
	printf(": avg:(%8.3g /%8.3g)= %8.3g <> maxerr: (%8.3g /%8.3g)= %8.3g", 
	err, sum, err/sum, maxerr, sum, maxerr/sum);
	if(nerr){
	    printf("\n%d errors detected \n");
	    exit(-2);
	}
	free(pa);
    }
    printf("\n");
    return(0);
}

int is_random;

void GetArguments( argc, argv)
int argc;
char *argv[];
{
    int i, j, k;
    int nerror = 0;

#define ON	1

    n_trials = DEF_TRIALS;
    is_random = 1;

/* ******************************************************* */
    for ( i = 1 ; (i < argc) && (nerror != ON) ; i ++ ) {
	if( argv[i][0] == '-') {
	    switch ( argv[i][1]) {
		case 'i' :
		case 'I' :  
			    is_random = 0;
			    break;
		default  :  nerror = ON;
	    }
	}
	else {
	    if( i+1 > argc)
		nerror = ON;
	    else { 
		n_trials = atoi( argv[i]);
	    }
	}
    }
    if( nerror == ON) {
	fprintf( stderr, 
	    "Usage : %s [-p(arallel)] [n_trials]\n", argv[0]);
	exit(-1);
    }
}

void get_values()
{
  if( is_random){ 
    size = random() % MAX_SIZE + 1;
    inc = random() % MAX_STRIDE + 1;
  } else{
    printf( "Enter SIZE : ");
    scanf("%d", &size);
    printf( "Enter STRIDE : ");
    scanf("%d", &inc);
  }
}
